"""Top-level module cli for dailylog."""

import logging
import sys
from types import MappingProxyType
from typing import NoReturn, Optional

import click
from click.core import Context

from dailylog1.cache import CONST_DAY, Cache
from dailylog1.config import Config
from dailylog1.constants import VERSION

CONTEXT_SETTINGS = MappingProxyType({"help_option_names": ["-h", "--help"]})
LOG_LEVELS = MappingProxyType(
    {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
    },
)


def log_label(level: str) -> str:
    """Return logger level name.

    Parameters
    ----------
    level : str
        Level number or name

    Returns
    -------
    str
        Level name, Default if not matched, by default "ERROR"
    """
    if level.isdigit():
        i_level = int(level)
        for key, valor in LOG_LEVELS.items():
            if i_level == valor:
                return key
    else:
        u_level = level.upper()
        if u_level in LOG_LEVELS:
            return u_level

    return "ERROR"


@click.command()
@click.option("--key", "-k", type=str, required=True, help="Specify key")
@click.option("--message", "-m", type=str, required=True, help="Specify message")
@click.option("--quiet/--no-quiet", "-q", default=False, help="Specify quiet flag")
@click.option(
    "--suppress",
    "-s",
    default=CONST_DAY,
    help="Specify seconds to suppress (default 86400 [one day])",
)
@click.option(
    "--level",
    "-l",
    default="ERROR",
    help="Specify one of CRITICAL, ERROR, WARNING, INFO, DEBUG (default: ERROR)",
)
@click.option(
    "--log_fn",
    "-f",
    type=str,
    required=False,
    help="Specify alternate log file",
)
@click.pass_context
def log(
    ctx: Context,
    key: str,
    message: str,
    quiet: bool,
    suppress: int,
    level: str,
    log_fn: Optional[str],
) -> NoReturn:
    """Log a message."""
    cache = Cache(ctx)
    if log_fn is None:
        log_fn = cache.default_log()
    cache.log_message(
        key,
        message,
        label=log_label(level),
        suppress=suppress,
        logfn=log_fn,
        quiet=quiet,
    )
    sys.exit(0)


@click.command()
@click.argument("log_fn", type=str, required=True, nargs=1)
@click.pass_context
def set_default_log(ctx: Context, log_fn: str) -> NoReturn:
    """Set a new default log."""
    config = Config(ctx)
    config.set_default_log(log_fn)
    if config.is_test():
        # For coverage purposes
        config.is_verbose()
    sys.exit(0)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-C",
    "--cache",
    required=False,
    type=str,
    help="specify alternate cache file (default ~/.cache/dailylog.json)",
)
@click.option(
    "-c",
    "--config",
    required=False,
    type=str,
    help="specify alternate config file (default ~/.config/dailylog.yaml)",
)
@click.option("-d", "--debug", count=True, default=0, help="increment debug level")
@click.option("-t", "--test/--no-test", default=False, help="specify test mode")
@click.option(
    "-v",
    "--verbose",
    count=True,
    default=0,
    help="increment verbosity level",
)
@click.version_option(VERSION)
@click.pass_context
def main(ctx, cache, config, debug, test, verbose):
    """Entry point for click script."""
    ctx.ensure_object(dict)
    if cache is None:
        cache = ""
    if config is None:
        config = ""
    ctx.obj["debug"] = debug
    ctx.obj["test"] = test
    ctx.obj["verbose"] = verbose
    ctx.obj["cache"] = cache
    ctx.obj["config"] = config


main.add_command(log)
main.add_command(set_default_log)

if __name__ == "__main__":
    sys.exit(main(obj={}))  # pragma no cover
# vim:ft=py noqa: E800



"""Top level module options for dailylog."""

from pathlib import Path

from click.core import Context
from wtforglib.dirs import ensure_directory


class Options:
    """Class to manage the options."""

    debug: int
    test: bool
    verbose: int
    cache_fn: str
    config_fn: str

    def __init__(self, ctx: Context) -> None:
        """Constructor for cache class.

        Args:
            ctx (Context): dictionary of click options

        Raises:
            ValueError: when config is an empty string
            ValueError: when config is not an absolute path
        """
        self.debug = int(ctx.obj.get("debug", 0))
        self.test = bool(ctx.obj.get("test", False))
        self.verbose = int(ctx.obj.get("verbose", 0))

        key = "cache"
        fn = str(ctx.obj.get(key, ""))
        if not fn:
            path = Path.home() / ".cache"
            ensure_directory(path)
            fn = str(path / "dailylog.json")
        self.cache_fn = Options.validate_fn_absolute(key, fn)
        key = "config"
        fn = str(ctx.obj.get(key, ""))
        if not fn:
            path = Path.home() / ".config"
            ensure_directory(path)
            fn = str(path / "dailylog.yaml")
        self.config_fn = Options.validate_fn_absolute(key, fn)

    def is_debug(self) -> bool:
        """Return True if debug option is greater than 0."""
        return self.debug > 0

    def is_test(self) -> bool:
        """Return True if test option is greater than 0."""
        return self.test

    def is_verbose(self) -> bool:
        """Return True if verbose option is greater than 0."""
        return self.verbose > 0

    def config_path(self) -> Path:
        """Return the config file path."""
        return Path(self.config_fn)

    def cache_path(self) -> Path:
        """Return the cache file path."""
        return Path(self.cache_fn)

    @staticmethod
    def validate_fn_absolute(file_key: str, file_name: str) -> str:
        """Validate an absolute file name/path.

        Args:
            file_key (str): key name of file "config | cache"
            file_name (str): path name of file to validate

        Raises:
            ValueError: when file_name is empty string
            ValueError: when file_name is not absolute

        Returns:
            str: validate file name/path as a string
        """
        if file_name == "":
            raise ValueError("{0} path name cannot be empty".format(file_key))
        path = Path(file_name)
        if not path.is_absolute():
            raise ValueError("{0} path name must be absolute".format(file_key))
        return str(path)



"""Top level module config for dailylog."""

import os
from pathlib import Path

import click
from click.core import Context
from wtforglib.files import load_yaml_file, write_yaml_file
from wtforglib.kinds import StrAnyDict

from dailylog1.exceptions import FilePermError
from dailylog1.options import Options

CURRENT_CONFIG_VERSION = 1
CONST_DEFAULT_LOG = str(Path.home() / "daily.log")


# WPS214 Found too many methods
class Config(Options):  # noqa: WPS214
    """Class to manage the configuration."""

    config: StrAnyDict

    def __init__(self, ctx: Context):
        """Constructor for Config class."""
        super().__init__(ctx)
        if self.config_path().is_file():
            self.config = self._load_config()
        else:
            self.config = {}
            self.config["version"] = CURRENT_CONFIG_VERSION
            self.config["default_log"] = CONST_DEFAULT_LOG
            self._save_config()

    def set_default_log(self, log_fn: str):
        """Set the default log file.

        Parameters
        ----------
        log_fn : str
            Absolute path to the log file
        """
        log_path = Path(log_fn)
        Config.validate_path(log_path)
        self.config["default_log"] = str(log_path)
        self._save_config()

    @staticmethod
    def update_config(config: StrAnyDict) -> StrAnyDict:
        """Update configuration from version to current version.

        Args:
            config (StrAnyDict): previous config verstion
        """
        version = config.get("version", 0)
        if version != CURRENT_CONFIG_VERSION:
            raise ValueError("Unknown config version: {0}".format(version))
        return config

    @staticmethod
    def validate_path(path: Path) -> None:
        """Validate the file path exists or can be created.

        Args:
            path (Path): files spec for log

        Raises:
            FileNotFoundError: when parent directory doesn't exist or is not a directory
            FilePermError: when parent directory is not writable
        """
        if path.exists():
            Config.validate_existing_path(path)
        elif not path.parent.is_dir():
            raise FileNotFoundError("Directory not found: {0}".format(path.parent))
        elif not os.access(path.parent, os.W_OK):
            raise FilePermError("Not writable: {0}".format(path.parent))

    def default_log(self) -> str:
        """Returns the path to the default log.

        Returns
        -------
        str
            Path to the default log
        """
        return self.config.get("default_log", CONST_DEFAULT_LOG)

    @staticmethod
    def validate_existing_path(path: Path) -> None:
        """Validate existing path.

        Args:
            path (Path): files spec for log

        Raises:
            FileNotFoundError: when path exists but is not a file
            FilePermError: when path exists but is not writable
        """
        if not path.is_file():
            raise FileNotFoundError("Not a file: {0}".format(path))
        if not os.access(path, os.W_OK):
            raise FilePermError("Not writable: {0}".format(path))

    def _load_config(self) -> StrAnyDict:
        """Load configuration from file."""
        config = load_yaml_file(self.config_path())
        version = config.get("version", 0)
        if version != CURRENT_CONFIG_VERSION:
            return Config.update_config(config)
        return config

    def _save_config(self) -> None:
        """Load configuration from file."""
        if self.is_debug():
            click.echo("Saving configuration to file: {0}".format(self.config_path()))
        write_yaml_file(self.config_path(), self.config)
